"""Visualizations for tradeoff analysis."""

from datetime import datetime
from micom.viz import Visualization
import numpy as np
import pandas as pd


def plot_tradeoff(
    tradeoff_rates,
    filename="tradeoff_%s.html" % datetime.now().strftime("%Y%m%d"),
):
    """Plot diagnostics for varying tradeoff values.

    Parameters
    ----------
    tradeoff_rates : pandas.DataFrame
        The growth rates returned by the `tradeoff` workflow.
    filename : str
        The HTML file where the visualization will be saved.

    Returns
    -------
    Visualization
        A MICOM visualization. Can be served with `viz.serve`.

    """
    rates = tradeoff_rates
    data = {"tradeoff": rates}
    viz = Visualization(filename, data, "tradeoff.html")
    growth = rates[
        ["taxon", "sample_id", "abundance", "tradeoff", "growth_rate", "tolerance"]
    ].copy()
    growth.tradeoff = growth.tradeoff.round(6).astype(str)
    growth.loc[growth.tradeoff == "nan", "tradeoff"] = "none"
    growth.loc[growth.growth_rate < growth.tolerance, "growth_rate"] = (
        growth.loc[growth.growth_rate < growth.tolerance, "tolerance"])
    growth.loc[:, "log_growth_rate"] = np.log10(growth.growth_rate)
    tradeoff = (
        growth.groupby(["tradeoff", "sample_id"])
        .apply(
            lambda df: pd.Series(
                {
                    "n_taxa": df.shape[0],
                    "n_growing": df[df.growth_rate > df.tolerance].shape[0],
                    "fraction_growing": (
                        df[df.growth_rate > df.tolerance].shape[0] / df.shape[0]
                    ),
                }
            )
        )
        .reset_index()
    )
    viz.save(
        growth=growth.to_json(orient="records"),
        tradeoff=tradeoff.to_json(orient="records"),
        extent=[growth.log_growth_rate.min(), growth.log_growth_rate.max()],
        width=400,
        height=300,
    )

    return viz



"""Build a database of organism metabolic models."""

import pandas as pd
import os
from os import path
from zipfile import ZipFile


def load_manifest(folder):
    """Get the manifest from a model DB."""
    mpath = path.join(folder, "manifest.csv")
    if not path.exists(mpath):
        raise ValueError(
            "No manifest found. `%s` does not look like a valid "
            "model database." % folder
        )
    manifest = pd.read_csv(path.join(folder, "manifest.csv"))
    if "file" not in manifest.columns:
        raise ValueError("Invalid manifest for model database :(")
    manifest.file = [path.join(folder, f) for f in manifest.file]
    return manifest


def load_zip_model_db(artifact, extract_path):
    """Prepare a model database for use."""
    if not path.exists(extract_path):
        os.mkdir(extract_path)
    with ZipFile(artifact) as zf:
        zf.extractall(extract_path)
    manifest = load_manifest(extract_path)
    manifest["file"] = [path.join(extract_path, f) for f in manifest.file]
    return manifest



"""Implements optimization and model problems."""

from micom.duality import fast_dual
from micom.util import (
    _format_min_growth,
    _apply_min_growth,
    check_modification,
)
from micom.logger import logger
from micom.solution import solve
from optlang.symbolics import Zero
from functools import partial


def add_dualized_optcom(community, min_growth):
    """Add dual Optcom variables and constraints to a community.

    Uses the original formulation of OptCom and solves the following
    multi-objective problem::

        maximize community_growth
        s.t. maximize growth_rate_i for all i
             s.t. Sv_i = 0
                  lb_i >= v_i >= ub_i

    Notes
    -----
    This method will only find one arbitrary solution from the Pareto front.
    There may exist several other optimal solutions.

    Arguments
    ---------
    community : micom.Community
        The community to modify.
    min_growth : positive float or array-like object.
        The minimum growth rate for each individual in the community. Either
        a single value applied to all individuals or one value for each.

    """
    logger.info("adding dual optcom to %s" % community.id)
    check_modification(community)
    min_growth = _format_min_growth(min_growth, community.taxa)

    prob = community.solver.interface

    # Temporarily subtitute objective with sum of individual objectives
    # for correct dual variables
    old_obj = community.objective
    community.objective = Zero
    for sp in community.taxa:
        taxa_obj = community.constraints["objective_" + sp]
        community.objective += taxa_obj.expression

    _apply_min_growth(community, min_growth)
    dual_coefs = fast_dual(community)

    logger.info("adding expressions for %d taxa" % len(community.taxa))
    for sp in community.taxa:
        primal_const = community.constraints["objective_" + sp]
        coefs = primal_const.get_linear_coefficients(primal_const.variables)
        coefs.update(
            {
                dual_var: -coef
                for dual_var, coef in dual_coefs.items()
                if sp in dual_var.name
            }
        )
        obj_constraint = prob.Constraint(
            Zero, lb=0, ub=0, name="optcom_suboptimality_" + sp
        )
        community.add_cons_vars([obj_constraint])
        community.solver.update()
        obj_constraint.set_linear_coefficients(coefs)

    community.objective = old_obj
    community.modification = "dual optcom"
    logger.info("finished adding dual optcom to %s" % community.id)


def add_moma_optcom(community, min_growth, linear=False):
    """Add a dualized MOMA version of OptCom.

    Solves a MOMA (minimization of metabolic adjustment) formulation of OptCom
    given by::

        minimize cooperativity_cost
        s.t. maximize community_objective
             s.t. Sv = 0
                  lb >= v >= ub
        where community_cost = sum (growth_rate - max_growth)**2
              if linear=False or
              community_cost = sum |growth_rate - max_growth|
              if linear=True

    Arguments
    ---------
    community : micom.Community
        The community to modify.
    min_growth : positive float or array-like object.
        The minimum growth rate for each individual in the community. Either
        a single value applied to all individuals or one value for each.
    linear : boolean
        Whether to use a non-linear (sum of squares) or linear version of the
        cooperativity cost. If set to False requires a QP-capable solver.

    """
    logger.info(
        "adding dual %s moma to %s"
        % ("linear" if linear else "quadratic", community.id)
    )
    check_modification(community)
    min_growth = _format_min_growth(min_growth, community.taxa)

    prob = community.solver.interface
    old_obj = community.objective
    coefs = old_obj.get_linear_coefficients(old_obj.variables)

    # Get maximum individual growth rates
    max_gcs = community.optimize_all(progress=False)

    _apply_min_growth(community, min_growth)
    dual_coefs = fast_dual(community)
    coefs.update({v: -coef for v, coef in dual_coefs.items()})
    obj_constraint = prob.Constraint(
        Zero, lb=0, ub=0, name="optcom_suboptimality"
    )
    community.add_cons_vars([obj_constraint])
    community.solver.update()
    obj_constraint.set_linear_coefficients(coefs)
    obj_expr = Zero
    logger.info("adding expressions for %d taxa" % len(community.taxa))
    for sp in community.taxa:
        v = prob.Variable("gc_constant_" + sp, lb=max_gcs[sp], ub=max_gcs[sp])
        community.add_cons_vars([v])
        taxa_obj = community.constraints["objective_" + sp]
        ex = v - taxa_obj.expression
        if not linear:
            ex = ex ** 2
        obj_expr += ex.expand()
    community.objective = prob.Objective(obj_expr, direction="min")
    community.modification = "moma optcom"
    logger.info("finished dual moma to %s" % community.id)


_methods = {
    "original": [add_dualized_optcom],
    "moma": [add_moma_optcom],
    "lmoma": [partial(add_moma_optcom, linear=True)],
}


def optcom(community, strategy, min_growth, fluxes, pfba):
    """Run OptCom for the community.

    OptCom methods are a group of optimization procedures to find community
    solutions that provide a tradeoff between the cooperative community
    growth and the egoistic growth of each individual [#p1]_. `micom`
    provides several strategies that can be used to find optimal solutions:

    - "moma": Minimization of metabolic adjustment. Simultaneously
      optimizes the community objective (maximize) and the cooperativity
      cost (minimize). This method finds an exact maximum but doubles the
      number of required variables, thus being slow.
    - "lmoma": The same as "moma" only with a linear
      representation of the cooperativity cost (absolute value).
    - "original": Solves the multi-objective problem described in [#p1]_.
      Here, the community growth rate is maximized simultanously with all
      individual growth rates. Note that there are usually many
      Pareto-optimal solutions to this problem and the method will only
      give one solution. This is also the slowest method.

    Parameters
    ----------
    community : micom.Community
        The community to optimize.
    strategy : str
        The strategy used to solve the OptCom formulation. Defaults to
        "lagrangian" which gives a decent tradeoff between speed and
        correctness.
    min_growth : float or array-like
        The minimal growth rate required for each individual. May be a
        single value or an array-like object with the same length as there
        are individuals.
    fluxes : boolean
        Whether to return the fluxes as well.
    pfba : boolean
        Whether to obtain fluxes by parsimonious FBA rather than
        "classical" FBA.

    Returns
    -------
    micom.CommunitySolution
        The solution of the optimization. If fluxes==False will only contain
        the objective value, community growth rate and individual growth rates.

    References
    ----------
    .. [#p1] OptCom: a multi-level optimization framework for the metabolic
       modeling and analysis of microbial communities.
       Zomorrodi AR, Maranas CD. PLoS Comput Biol. 2012 Feb;8(2):e1002363.
       doi: 10.1371/journal.pcbi.1002363, PMID: 22319433

    """
    if strategy not in _methods:
        raise ValueError(
            "strategy must be one of {}!".format(",".join(_methods))
        )
    funcs = _methods[strategy]

    with community as com:
        # Add needed variables etc.
        funcs[0](com, min_growth)
        return solve(community, fluxes=fluxes, pfba=pfba)



"""Visualization for growth rates."""

from datetime import datetime
from micom.viz import Visualization


def plot_growth(
    results,
    filename="growth_rates_%s.html" % datetime.now().strftime("%Y%m%d")
):
    """Plot the taxa growth rates.

    Parameters
    ----------
    results : micom.workflows.GrowthResults
        The results returned by the `grow` workflow.
    filename : str
        The HTML file where the visualization will be saved.

    Returns
    -------
    Visualization
        A MICOM visualization. Can be served with `viz.serve`.
    """
    rates = results.growth_rates
    rates = rates[rates.growth_rate > rates.tolerance][
        ["taxon", "sample_id", "abundance", "growth_rate"]
    ]
    data = {"growth_rates": rates}
    viz = Visualization(filename, data, "growth.html")
    viz.save(
        data=rates.to_json(orient="records"), width=800, height=400
    )

    return viz



"""Helper to convert external data to a micom taxonomy."""

from micom.db import load_manifest, load_zip_model_db
from micom.community import _ranks
from micom.qiime_formats import (
    load_qiime_model_db,
    load_qiime_feature_table,
    load_qiime_taxonomy,
)
import pandas as pd
from tempfile import TemporaryDirectory


def build_from_qiime(
    abundance,
    taxonomy: pd.Series,
    manifest: pd.DataFrame,
    cutoff: float = 1e-4,
    strict: bool = True,
) -> pd.DataFrame:
    """Build the specification for the community models."""
    taxa = taxonomy.str.replace("[\\w_]+__|\\[|\\]", "")
    taxa = taxa.str.split(";\\s*", expand=True).replace("", None)
    taxa.columns = _ranks[0 : taxa.shape[1]]
    taxa["taxid"] = taxonomy.index
    taxa.index == taxa.taxid

    rank = manifest.summary_rank[0]
    if strict:
        ranks = [
            r
            for r in _ranks[0 : (_ranks.index(rank) + 1)]
            if r in taxa.columns and r in manifest.columns
        ]
    else:
        ranks = [rank]
    taxa["mapping_ranks"] = taxa[ranks].apply(
        lambda s: "|".join(s.astype("str")), axis=1
    )

    abundance = (
        abundance.collapse(
            lambda id_, x: taxa.loc[id_, "mapping_ranks"],
            axis="observation", norm=False
        )
        .to_dataframe(dense=True)
        .T
    )
    abundance["sample_id"] = abundance.index

    abundance = abundance.melt(
        id_vars="sample_id", var_name="mapping_ranks", value_name="abundance"
    )
    abundance = pd.merge(
        abundance, taxa[ranks + ["mapping_ranks"]].drop_duplicates(), on="mapping_ranks"
    )
    del abundance["mapping_ranks"]
    depth = abundance.groupby("sample_id").abundance.sum()
    abundance["relative"] = abundance.abundance / depth[abundance.sample_id].values
    abundance.dropna(subset=ranks, inplace=True)

    micom_taxonomy = pd.merge(manifest, abundance, on=ranks)
    micom_taxonomy = micom_taxonomy[micom_taxonomy.relative > cutoff]
    print("Taxa per sample:")
    print(micom_taxonomy.sample_id.value_counts().describe(), "\n")
    return abundance


def qiime_to_micom(feature_table, taxonomy, model_db, cutoff, strict=False):
    """Load a micom taxonomy from Qiime 2 data.

    Parameters
    ----------
    feature_table : str
        Path to a Qiime 2 FeatureTable artifact.
    taxonomy : str
        Path to a Qiime 2 FeatureData[Taxonomy] artifact.
    model_db : str
        Path to a model database folder.
    cutoff : float
        Minimum relative abundance considered present in a sample.
    strict : bool
        Whether to match taxa on all ranks or only the target taxonomy rank.

    Returns
    -------
    pd.DataFrame
        A micom taxonomy containing abundances and taxonomy calls in long
        format.
    """
    compressed = model_db.endswith(".qza") or model_db.endswith(".zip")
    if compressed:
        tdir = TemporaryDirectory(prefix="micom_")
    if model_db.endswith(".qza"):
        manifest = load_qiime_model_db(model_db, tdir.name)
    elif model_db.endswith(".zip"):
        manifest = load_zip_model_db(model_db, tdir.name)
    else:
        manifest = load_manifest(model_db)

    table = load_qiime_feature_table(feature_table)
    taxonomy = load_qiime_taxonomy(taxonomy)

    return build_from_qiime(table, taxonomy, manifest, cutoff, strict)



"""Visualization for exchanges."""

from datetime import datetime
from fastcluster import linkage
from scipy.cluster.hierarchy import leaves_list
from micom.viz.core import Visualization
import pandas as pd
from sklearn.manifold import TSNE


def plot_exchanges_per_sample(
    results,
    filename="sample_exchanges_%s.html" % datetime.now().strftime("%Y%m%d"),
    direction="import",
    cluster=True,
) -> None:
    """Plot the per sample exchange fluxes.

    Parameters
    ----------
    results : micom.workflows.GrowthResults
        The results returned by the `grow` workflow.
    filename : str
        The HTML file where the visualization will be saved.
    direction : str either "import" or "export"
        The direction of fluxes to plot.
    cluster : bool
        Whether to reorder samples so that samples with similar exchange
        fluxes are close to another.

    Returns
    -------
    Visualization
        A MICOM visualization. Can be served with `viz.serve`.
    """
    exchanges = results.exchanges
    anns = results.annotations
    anns.index = anns.metabolite
    if direction not in ["import", "export"]:
        ValueError("Not a valid flux direction. Must be `import` or `export`.")
    exchanges = exchanges[
        (exchanges.taxon == "medium")
        & (exchanges.direction == direction)
        & (exchanges.flux.abs() > 1e-6)
    ].copy()
    exchanges.flux = exchanges.flux.abs()
    mat = exchanges.pivot_table(
        values="flux", index="metabolite", columns="sample_id", fill_value=1e-6
    )
    sample_order = leaves_list(linkage(mat.values.T, method="average"))
    reaction_order = leaves_list(linkage(mat.values, method="average"))
    mat = mat.iloc[reaction_order, sample_order]

    mat["metabolite"] = mat.index
    data = mat.melt(
        id_vars="metabolite", var_name="sample_id", value_name="flux"
    )
    data["description"] = anns.loc[data.metabolite, "name"].values
    data = {"exchange_fluxes": data}
    viz = Visualization(filename, data, "sample_heatmap.html")
    w = mat.shape[1] * 10
    viz.save(
        data=data["exchange_fluxes"].to_json(orient="records"),
        width=w,
        height=w * mat.shape[0] / mat.shape[1],
    )
    return viz


def plot_exchanges_per_taxon(
    results,
    filename="taxon_exchanges_%s.html" % datetime.now().strftime("%Y%m%d"),
    direction="import",
    **tsne_args
) -> None:
    """Plot the exchange fluxes per taxon.

    Parameters
    ----------
    results : micom.workflows.GrowthResults
        The exchanges returned by the `grow` workflow.
    filename : str
        The HTML file where the visualization will be saved.
    direction : str either "import" or "export"
        The direction of fluxes to plot.
    tsne_args : dict
        Additional arguments passed to TSNE.

    Returns
    -------
    Visualization
        A MICOM visualization. Can be served with `viz.serve`.

    """
    exchanges = results.exchanges
    if direction not in ["import", "export"]:
        ValueError("Not a valid flux direction. Must be `import` or `export`.")
    exchanges = exchanges[
        (exchanges.taxon != "medium") & (exchanges.direction == direction)
    ].copy()
    exchanges["flux"] = exchanges.flux.abs() * exchanges.abundance
    mat = exchanges.pivot_table(
        values="flux",
        index=["sample_id", "taxon"],
        columns="reaction",
        fill_value=0,
    )
    reduced = TSNE(**tsne_args).fit_transform(mat.values)
    reduced = pd.DataFrame(
        reduced, index=mat.index, columns=["TSNE 1", "TSNE 2"]
    ).reset_index()
    data = {"reduced": reduced}
    viz = Visualization(filename, data, "reduced.html")
    viz.save(data=reduced.to_json(orient="records"), width=600, height=500)

    return viz



"""Makes it easier to run analyses on several samples in parallel."""

from collections import Sized, namedtuple
from multiprocessing import Pool
from tqdm.auto import tqdm

GrowthResults = namedtuple(
    "GrowthResults",
    ["growth_rates", "exchanges", "annotations"]
)


def workflow(func, args, n_jobs=4, unit="sample(s)", progress=True):
    """Run analyses for several samples in parallel.

    This will analyze several samples in parallel. Includes a workaround for
    optlang memory leak.

    Arguments
    ---------
    func : function
        A function that takes a single argument (can be any object) and
        that performs your analysis for a single sample.
    args : array-like object
        An array-like object (list, tuple, numpy array, pandas Series, etc.)
        that contains the arguments for each sample.
    n_jobs : positive int
        How many samples to analyze in parallel at once.
    unit : str
        The unit used for the progress bar.
    progress : bool
        Whether to show a progress bar.
    """
    if not isinstance(args, Sized):
        ValueError("`args` must have a length.")

    with Pool(processes=n_jobs, maxtasksperchild=1) as pool:
        it = pool.imap_unordered(func, args)
        if progress:
            it = tqdm(it, total=len(args), unit=unit)
        results = list(it)
    return results



"""setup.py for micom."""

from setuptools import setup, find_packages

# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="micom",
    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version="0.18.0",
    description="Microbial community modeling based on cobrapy.",
    long_description=long_description,
    # The project's main homepage.
    url="https://github.com/micom-dev/micom",
    # Author details
    author="Christian Diener",
    author_email="mail@cdiener.com",
    # Choose your license
    license="Apache License 2.0",
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 5 - Production/Stable",
        # Indicate who your project is intended for
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        # Pick your license as you wish (should match "license" above)
        "License :: OSI Approved :: Apache Software License",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    # What does your project relate to?
    keywords="microbiota modeling metabolism community",
    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=["docs", "tests"]),
    package_data={
        "micom": [
            path.join("data", "*.csv"),
            path.join("data", "*.gz"),
            path.join("data", "templates", "*.*"),
            path.join("data", "artifacts", "*.qza")
        ]
    },
    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],
    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        "cobra>=0.17.1",
        "optlang>=1.4.4",
        "pandas>=0.20.0",
        "tqdm>=4.14",
        "loguru>=0.3.2",
        "jinja2>=2.10.0",
        "scikit-learn>=0.22.0",
        "scipy>=1.0.0",
        "fastcluster>=1.1.0",
        "symengine>=0.6.1"
    ],
    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        "dev": [""],
        "test": ["coverage", "pytest", "pytest-cov"],
    },
)
